local entity_id = GetUpdatedEntityID()
local root_id = EntityGetRootEntity( entity_id )
local rx, ry = EntityGetTransform(root_id)
-- EntitySetTransform(entity_id, rx, ry - 20)

local function get_state()
	local entity_id = GetUpdatedEntityID()
	_state = _state or {}
	_state[entity_id] = _state[entity_id] or {}
	for i, v in ipairs(_state) do
		if not EntityGetIsAlive(v) then
			-- Remove saved states of entities that no longer exist
			-- to prevent memory leak
			_state[entity_id] = nil
		end
	end
	return _state[entity_id]
end

local function distance(x1, y1, x2, y2)
  local dx = x2 - x1
  local dy = y2 - y1
  return math.sqrt(dx * dx + dy * dy)
end
-- Function to subdivide a 2D segmented line into parts of equal length
local function subdivide_line(segmented_line, equal_length)
  if #segmented_line < 2 or equal_length <= 0 then
      return nil
  end

  local total_length = 0
  for i = 1, #segmented_line - 1 do
      local x1, y1 = segmented_line[i].x, segmented_line[i].y
      local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
      total_length = total_length + distance(x1, y1, x2, y2)
  end

  local num_subdivisions = math.floor(total_length / equal_length)
  if num_subdivisions == 0 then
      return nil
  end

  local segment_length = total_length / num_subdivisions

  local result = {}
  table.insert(result, segmented_line[1]) -- Include the starting point

  local current_length = 0
  for i = 1, #segmented_line - 1 do
      local x1, y1 = segmented_line[i].x, segmented_line[i].y
      local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
      local segment_dist = distance(x1, y1, x2, y2)

      while current_length + segment_dist >= segment_length do
          local t = (segment_length - current_length) / segment_dist
          local new_x = x1 + t * (x2 - x1)
          local new_y = y1 + t * (y2 - y1)
          table.insert(result, { x = new_x, y = new_y })
          segment_dist = segment_dist - (segment_length - current_length)
          x1, y1 = new_x, new_y
          current_length = 0
      end

      current_length = current_length + segment_dist
  end

  -- Include the ending point
  table.insert(result, segmented_line[#segmented_line])

  return result
end

local function map(value, old_min, old_max, new_min, new_max)
	return (value - old_min) * (new_max - new_min) / (old_max - old_min) + new_min;
end


-- ################################
-- ### END OF UTILITY FUNCTIONS ###
-- ################################

local x, y, rot, scale_x = EntityGetTransform(entity_id)
local state = get_state()
state.path = state.path or {}

local sprite_comp = EntityGetFirstComponentIncludingDisabled(root_id, "SpriteComponent")
ComponentSetValue2(sprite_comp, "rect_animation", "mouth_open")
ComponentSetValue2(sprite_comp, "next_rect_animation", "mouth_open")

local child_found = false
local proj_x, proj_y
local child_entities = EntityGetAllChildren(entity_id) or {}
for i, v in ipairs(child_entities) do
  if EntityGetName(v) == "tongue_projectile" then
	proj_x, proj_y = EntityGetTransform(v)
	child_found = true
	break
  end
end

-- Construct or deconstruct the points for the path
-- and pull enemies or damage them
if child_found == true then
	table.insert(state.path, { x = proj_x, y = proj_y })
else
	if #state.path > 0 then
		-- Tongue is retracting
		local range = 15
		-- locate nearest enemy
		local min_dist = 15
		local tip = state.path[#state.path]
		local tip = { x = tip.x + (rx - x), y = tip.y + (ry - y - 6) }
		for _,id in pairs(EntityGetInRadiusWithTag(tip.x, tip.y, range, "mortal")) do
			-- is target a valid enemy
			if EntityGetComponent(id, "GenomeDataComponent") ~= nil and EntityGetHerdRelation(root_id, id) < 40 and ( id ~= entity_id ) then
				local ex, ey = EntityGetFirstHitboxCenter( id )
				local dist = distance(tip.x, tip.y, ex, ey)
				if dist < min_dist then
					EntitySetTransform(id, tip.x, tip.y)
				end
				break
			end
		end
		table.remove(state.path, #state.path)
	else
		-- Damage
		local range = 25
		for _,id in pairs(EntityGetInRadiusWithTag(x, y, range, "mortal")) do
			-- is target a valid enemy
			if EntityGetComponent(id, "GenomeDataComponent") ~= nil and EntityGetHerdRelation(root_id, id) < 40 and ( id ~= entity_id ) then
				local ex, ey = EntityGetFirstHitboxCenter( id )
				local dist = distance(x, y, ex, ey)
				if dist < range then
					if ( id > 0 ) and (id ~= nil) then
						local orbcount = 0
							
						-- Get the amount of orbs to adjust boss difficulty
						orbcount = GameGetOrbCountThisRun()

						if orbcount > 15 then
							local value = ( ( ( 2.0 ^ (15 + 1.3) ) + (15*15.5) ) * 0.0001 ) * 0.5
							boss_hp = value * ( orbcount - 15 )
						else
							boss_hp = ( ( ( 2.0 ^ (orbcount + 1.3) ) + (orbcount*15.5) ) * 0.0001 ) * 0.5
						end

						if boss_hp > 1000 then
							boss_hp = 1000
						end
						EntityInflictDamage( id, 0.5 + boss_hp, "DAMAGE_OVEREATING", "Swallow", "BLOOD_EXPLOSION", 0, 0, root_id, ex, ey, 0)
						EntityInflictDamage( id, 0.5 + boss_hp, "DAMAGE_MELEE", "Swallow", "BLOOD_EXPLOSION", 0, 0, root_id, ex, ey, 0)
					end

				end
			end
		end
		state[entity_id] = nil
		EntityKill(entity_id)
	end
end

-- Caculate the length of our tongue so we can scale it accordingly

local seg_distances = {}
local path_length = 0
for i=1, #state.path-1 do
  local p1 = state.path[i]
  local p2 = state.path[i+1]
  local dist = distance(p1.x, p1.y, p2.x, p2.y)
  path_length = path_length + dist
  if seg_distances[i] then
    seg_distances[i] = seg_distances[i-1] + dist
  else
    seg_distances[i] = dist
  end
end

-- Calculate a smooth path 
local max_length = 300
local dist_p = math.min(1, path_length / max_length)
dist_p = map(dist_p, 0, 1, 3.5, 2)
local new_path = subdivide_line(state.path, dist_p) or {}
local half_point = #new_path / 2
local sprite_size = 9

-- Create or remove SpriteComponents until it matches the amount of points we have on the path
local sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}
-- Create new comps until we have enough
for i=1, #new_path - #sprite_comps do
	EntityAddComponent2(entity_id, "SpriteComponent", {
		image_file="data/enemies_gfx/cyclops_slime_tongue_sprite.png",
		update_transform_rotation=false,
		has_special_scale=true,
		z_index=-1.1
	})
end
-- Remove comps that are too many
for i=1, #sprite_comps - #new_path do
	EntityRemoveComponent(entity_id, sprite_comps[i])
end
sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}

-- Set SpriteComponent offsets
for i, v in ipairs(new_path) do
	local h = (i - half_point) / #new_path
	h = h * h
	local scale = 1 - math.min(max_length, path_length) / max_length
	scale = scale + h * 2
	scale = math.min(1, math.max(0.4, scale))

	ComponentSetValue2(sprite_comps[i], "offset_x", -(v.x - x + (rx - x)) * (1 / scale) + (sprite_size/2))
    ComponentSetValue2(sprite_comps[i], "offset_y", -(v.y - y + (ry - y - 6)) * (1 / scale) + (sprite_size/2))
	ComponentSetValue2(sprite_comps[i], "special_scale_x", scale)
	ComponentSetValue2(sprite_comps[i], "special_scale_y", scale)
end

-- local entity_id = GetUpdatedEntityID()
-- local root_id = EntityGetRootEntity( entity_id )
-- local rx, ry, rrot = EntityGetTransform(root_id)

-- EntitySetTransform(entity_id, rx, ry - 6)
-- local x, y, rot = EntityGetTransform(entity_id)

-- local r_sprite_comp = EntityGetFirstComponentIncludingDisabled(root_id, "SpriteComponent") or {}
-- ComponentSetValue2(r_sprite_comp, "rect_animation", "mouth_open")
-- ComponentSetValue2(r_sprite_comp, "next_rect_animation", "mouth_open")

-- SetRandomSeed(x + GameGetFrameNum(), y)

-- local child_found = false

-- local child_entities = EntityGetAllChildren(entity_id) or {}
-- for i, v in ipairs(child_entities) do
  -- if EntityGetName(v) == "tongue_projectile" then
	-- tongue_tip_entity = v
	-- old_x = GetValueNumber("prev_x", x)
	-- old_y = GetValueNumber("prev_y", y)
	-- cx, cy, rot = EntityGetTransform(tongue_tip_entity)
	-- SetValueNumber("prev_x", cx)
	-- SetValueNumber("prev_y", cy)
	-- child_found = true
	-- break
  -- end
-- end

-- if child_found == true then
	
	-- -- Calculate the direction towards the target
	-- -- local dir_x = cx - old_x
	-- -- local dir_y = cy - old_y
	-- -- local dir_length = math.sqrt(dir_x * dir_x + dir_y * dir_y)

	-- -- -- Calculate the number of explosions to spawn
	-- -- local explosion_interval = 3  -- Adjust the interval between explosions
	-- -- local num_explosions = math.floor(dir_length / explosion_interval)

	-- -- -- Calculate the step size for each explosion
	-- -- local step_x = dir_x / num_explosions
	-- -- local step_y = dir_y / num_explosions

	-- -- -- Spawn explosions along the trajectory
	-- -- for i = 1, num_explosions do
		-- -- eoffset_x = step_x * i
		-- -- eoffset_y = step_y * i
		
		-- function lerp(a, b, weight)
			-- return a * weight + b * (1 - weight)
		-- end
		-- local steps=4
		-- for i=1, steps do
			-- local cur_x = lerp(cx, old_x, i / steps)
			-- local cur_y = lerp(cy, old_y, i / steps)
			-- number = GetValueNumber("number", 0)
			-- SetValueNumber("number", number + 0.00001)
			
			-- local x_value = (x - cur_x) + 5
			-- local y_value = (y - cur_y) + 5

			-- EntityAddComponent2(entity_id, "SpriteComponent", {
				-- image_file="data/enemies_gfx/chameleon_tongue_sprite.png",
				-- offset_x=x_value,
				-- offset_y=y_value,
				-- z_index= 0.61 - number,
				-- update_transform_rotation=false,
				-- has_special_scale=true,
				-- text=tostring(x_value)..","..tostring(y_value)
			-- })
		-- end
		
		-- local sprite_comps = EntityGetComponent(entity_id, "SpriteComponent") or {}
		-- for i2, comp in ipairs(sprite_comps) do
			-- local max_length = 100
			-- local scale = 1 - math.min(max_length, #sprite_comps) / max_length
			-- local h = (i2 - ( #sprite_comps / 2 )) / #sprite_comps
			-- local h = h * h
			-- local scale = scale + h * 2
			-- local scale = math.min(1, math.max(0.6, scale))
			
			-- ComponentSetValue2(comp, "special_scale_x", scale)
			-- ComponentSetValue2(comp, "special_scale_y", scale)
			-- local text_val = ComponentGetValue2(comp, "text")
			-- local x_str, y_str = text_val:match("([^,]+),([^,]+)")
			-- local offx,offy = tonumber(x_str), tonumber(y_str)
			-- ComponentSetValue2(comp, "offset_x", ( offx * scale) * (1 / scale))
			-- ComponentSetValue2(comp, "offset_y", ( offy * scale) * (1 / scale))
		-- end
		
		-- local range = 15
		
		-- local pos_x = x - cx
		-- local pos_y = y - cy
		-- -- locate nearest enemy
		-- local enemy, enemy_x, enemy_y
		-- local min_dist = 15
		-- for _,id in pairs(EntityGetInRadiusWithTag(pos_x, pos_y, range, "mortal")) do
			-- -- is target a valid enemy
			-- if EntityGetComponent(id, "GenomeDataComponent") ~= nil and EntityGetHerdRelation(root_id, id) < 40 and ( id ~= entity_id ) then
				-- local ex, ey = EntityGetFirstHitboxCenter( id )
				-- local dist = get_distance(pos_x, pos_y, ex, ey)
				-- if dist < min_dist then
					-- EntitySetTransform(id, pos_x, pos_y)
				-- end
			-- end
		-- end
	-- -- end
-- else
	
	-- local sprite_comps = EntityGetComponent(entity_id, "SpriteComponent") or {}
	-- local sprite_comp_table = {}
	-- for i2, comp in ipairs(sprite_comps) do
	  -- sprite_comp_table[i2] = {
		-- comp = comp,
		-- z_index = ComponentGetValue2(comp, "z_index")
	  -- }
	-- end
	-- table.sort(sprite_comp_table, function(a, b) return a.z_index > b.z_index end)
	-- if #sprite_comp_table > 0 then
		-- for i, v in ipairs(sprite_comp_table) do
			-- valuable_sprite_comp = v.comp
			-- -- valuable_sprite_comp = sprite_comp_table[1]
			
			-- local offx = ComponentGetValue2(valuable_sprite_comp, "offset_x")
			-- local offy = ComponentGetValue2(valuable_sprite_comp, "offset_y")
			-- pos_x = x - offx
			-- pos_y = y - offy
		-- end
		
		-- EntityRemoveComponent( entity_id, valuable_sprite_comp )
		
		-- local sprite_comps = EntityGetComponent(entity_id, "SpriteComponent") or {}
		-- for i2, comp in ipairs(sprite_comps) do
			-- local max_length = 99999
			-- local scale = 1 - math.min(max_length, #sprite_comps) / max_length
			-- local h = (i2 - ( #sprite_comps / 2 )) / #sprite_comps
			-- local h = h * h
			-- local scale = scale + h * 2
			-- local scale = math.min(1, math.max(0.6, scale))
			
			-- ComponentSetValue2(comp, "special_scale_x", scale)
			-- ComponentSetValue2(comp, "special_scale_y", scale)
			-- local text_val = ComponentGetValue2(comp, "text")
			-- local x_str, y_str = text_val:match("([^,]+),([^,]+)")
			-- local offx,offy = tonumber(x_str), tonumber(y_str)
			-- ComponentSetValue2(comp, "offset_x", ( ( offx ) * scale) * (1 / scale))
			-- ComponentSetValue2(comp, "offset_y", ( ( offy ) * scale) * (1 / scale))
		-- end
		
		-- local range = 15

		-- -- locate nearest enemy
		-- local enemy, enemy_x, enemy_y
		-- local min_dist = 15
		-- for _,id in pairs(EntityGetInRadiusWithTag(pos_x, pos_y, range, "mortal")) do
			-- -- is target a valid enemy
			-- if EntityGetComponent(id, "GenomeDataComponent") ~= nil and EntityGetHerdRelation(root_id, id) < 40 and ( id ~= entity_id ) then
				-- local ex, ey = EntityGetFirstHitboxCenter( id )
				-- local dist = get_distance(pos_x, pos_y, ex, ey)
				-- if dist < min_dist then
					-- EntitySetTransform(id, pos_x, pos_y)
				-- end
			-- end
		-- end
	-- else
		-- local range = 25

		-- -- locate nearest enemy
		-- local enemy, enemy_x, enemy_y
		-- local min_dist = 25
		-- for _,id in pairs(EntityGetInRadiusWithTag(x, y, range, "mortal")) do
			-- -- is target a valid enemy
			-- if EntityGetComponent(id, "GenomeDataComponent") ~= nil and EntityGetHerdRelation(root_id, id) < 40 and ( id ~= entity_id ) then
				-- local ex, ey = EntityGetFirstHitboxCenter( id )

				-- local distance = get_distance(x, y, ex, ey)

				-- if distance < range then
					-- if ( id > 0 ) and (id ~= nil) then
						-- EntityInflictDamage( id, 0.5, "DAMAGE_OVEREATING", "Swallow", "BLOOD_EXPLOSION", 0, 0, root_id, ex, ey, 0)
						-- EntityInflictDamage( id, 0.5, "DAMAGE_MELEE", "Swallow", "BLOOD_EXPLOSION", 0, 0, root_id, ex, ey, 0)
					-- end
					
				-- end
			-- end
		-- end
		-- EntityKill(entity_id)
	-- end
-- end