-- Constants for boid behavior
local MAX_SPEED = 230
local MAX_FORCE = 0.1
local NEIGHBOR_RADIUS = 50.0
local SEPARATION_DISTANCE = 50
local AVOIDANCE_DISTANCE = 40.0
local WANDER_FORCE = 400 -- Adjust the strength of the random wandering force

-- Retrieve the necessary components
local entity_id = GetUpdatedEntityID()
local velocity_component = EntityGetFirstComponentIncludingDisabled(entity_id, "VelocityComponent")

-- Get the current position and velocity of the entity
local x, y = EntityGetTransform(entity_id)
local vel_x, vel_y = ComponentGetValue2(velocity_component, "mVelocity")

-- Function to calculate the normalized vector
local function Normalize(x, y)
  local length = math.sqrt(x * x + y * y)
  return x / length, y / length
end

-- Function to limit the magnitude of a vector
local function LimitMagnitude(x, y, max_magnitude)
  local magnitude = math.sqrt(x * x + y * y)
  if magnitude > max_magnitude then
    return x * max_magnitude / magnitude, y * max_magnitude / magnitude
  end
  return x, y
end

-- Function to calculate the distance between two entities
local function Distance(entity_id1, entity_id2)
  local x1, y1 = EntityGetTransform(entity_id1)
  local x2, y2 = EntityGetTransform(entity_id2)
  local dx = x2 - x1
  local dy = y2 - y1
  return math.sqrt(dx * dx + dy * dy)
end

-- Function to calculate separation force from nearby entities
local function Separation(entity_id)
  local separation_force_x, separation_force_y = 0, 0
  local entities = EntityGetWithTag("projectile") -- Replace "boid" with the tag you assign to your boid entities

  for _, other_entity_id in ipairs(entities) do
    if other_entity_id ~= entity_id then
      local distance = Distance(entity_id, other_entity_id)
      if distance < SEPARATION_DISTANCE then
        local other_x, other_y = EntityGetTransform(other_entity_id)
        local dx = x - other_x
        local dy = y - other_y
        separation_force_x = separation_force_x + dx / distance
        separation_force_y = separation_force_y + dy / distance
      end
    end
  end

  return separation_force_x, separation_force_y
end

-- Function to calculate alignment force towards nearby entities
local function Alignment(entity_id)
  local alignment_force_x, alignment_force_y = 0, 0
  local entities = EntityGetWithTag("projectile") -- Replace "boid" with the tag you assign to your boid entities
  local neighbor_count = 0

  for _, other_entity_id in ipairs(entities) do
    if other_entity_id ~= entity_id then
      local distance = Distance(entity_id, other_entity_id)
      if distance < NEIGHBOR_RADIUS then
        local other_vel_x, other_vel_y = ComponentGetValue2(EntityGetFirstComponentIncludingDisabled(other_entity_id, "VelocityComponent"), "mVelocity")
        alignment_force_x = alignment_force_x + other_vel_x
        alignment_force_y = alignment_force_y + other_vel_y
        neighbor_count = neighbor_count + 1
      end
    end
  end

  if neighbor_count > 0 then
    alignment_force_x = alignment_force_x / neighbor_count
    alignment_force_y = alignment_force_y / neighbor_count
  end

  return alignment_force_x, alignment_force_y
end

-- Function to calculate cohesion force towards nearby entities
local function Cohesion(entity_id)
  local cohesion_force_x, cohesion_force_y = 0, 0
  local entities = EntityGetWithTag("projectile") -- Replace "boid" with the tag you assign to your boid entities
  local neighbor_count = 0

  for _, other_entity_id in ipairs(entities) do
    if other_entity_id ~= entity_id then
      local distance = Distance(entity_id, other_entity_id)
      if distance < NEIGHBOR_RADIUS then
        local other_x, other_y = EntityGetTransform(other_entity_id)
        cohesion_force_x = cohesion_force_x + other_x
        cohesion_force_y = cohesion_force_y + other_y
        neighbor_count = neighbor_count + 1
      end
    end
  end

  if neighbor_count > 0 then
    cohesion_force_x = cohesion_force_x / neighbor_count - x
    cohesion_force_y = cohesion_force_y / neighbor_count - y
  end

  return cohesion_force_x, cohesion_force_y
end

-- Function to avoid walls using surface normals
local function AvoidWalls()
  local avoidance_force_x, avoidance_force_y = 0, 0
  local ray_length = AVOIDANCE_DISTANCE
  local ray_count = 8 -- Adjust the number of rays to use for wall avoidance

  for i = 1, ray_count do
    local angle = (2 * math.pi * i) / ray_count
    local ray_x = x + math.cos(angle) * ray_length
    local ray_y = y + math.sin(angle) * ray_length

    local found_normal, normal_x, normal_y, approximate_distance = GetSurfaceNormal(ray_x, ray_y, ray_length, ray_count)

    if found_normal then
      local did_hit, hit_x, hit_y = RaytraceSurfaces(x, y, ray_x, ray_y)

      if did_hit then
        local avoidance_vector_x = hit_x - x
        local avoidance_vector_y = hit_y - y

        avoidance_force_x = avoidance_force_x - avoidance_vector_x
        avoidance_force_y = avoidance_force_y - avoidance_vector_y
      end
    end
  end

  return avoidance_force_x, avoidance_force_y
end

-- Function to calculate random wandering force
local function Wander()
  local wander_force_x = math.random() - 0.5
  local wander_force_y = math.random() - 0.5
  return wander_force_x, wander_force_y
end

-- Calculate the boid forces
local separation_force_x, separation_force_y = Separation(entity_id)
local alignment_force_x, alignment_force_y = Alignment(entity_id)
local cohesion_force_x, cohesion_force_y = Cohesion(entity_id)
local avoidance_force_x, avoidance_force_y = AvoidWalls()
local wander_force_x, wander_force_y = Wander()

-- Apply the forces to the velocity
vel_x = vel_x + separation_force_x + alignment_force_x + cohesion_force_x + avoidance_force_x + wander_force_x * WANDER_FORCE
vel_y = vel_y + separation_force_y + alignment_force_y + cohesion_force_y + avoidance_force_y + wander_force_y * WANDER_FORCE

-- Limit the velocity to the maximum speed
vel_x, vel_y = LimitMagnitude(vel_x, vel_y, MAX_SPEED)

-- Apply the new velocity to the entity
ComponentSetValue2(velocity_component, "mVelocity", vel_x, vel_y)

--chirping audio
-- local chirping_delay = GetValueNumber("delay", 30)
-- if ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted") % chirping_delay == 0 then
	-- -- GamePlaySound("data/audio/Desktop/animals.bank", "animals/rat/damage/projectile", x, y)
	-- GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/bird_chirp", x, y)
	-- SetValueNumber("delay",Random(60, 120))
-- end